#include <linux/types.h>
#include <linux/mm.h>

#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/bootinfo.h>
#include <asm/amigaints.h>
#include <asm/amigahw.h>
#include <asm/zorro.h>
#include <asm/irq.h>

#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "wd33c93.h"
#include "gvp11.h"

#define DMA(ptr) ((gvp11_scsiregs *)((ptr)->base))
#define HDATA(ptr) ((struct WD33C93_hostdata *)((ptr)->hostdata))

static struct Scsi_Host *first_instance = NULL;
static Scsi_Host_Template *gvp11_template;

static void gvp11_intr (int irq, struct intframe *fp, void *data)
{
    unsigned int status;
    struct Scsi_Host *instance;

    for (instance = first_instance; instance &&
	 instance->hostt == gvp11_template; instance = instance->next)
    {
	status = DMA(instance)->CNTR;
	if (!(status & GVP11_DMAC_INT_PENDING))
	    continue;

	/* disable PORTS interrupt */
	custom.intena = IF_PORTS;
	wd33c93_intr (instance);
	/* enable PORTS interrupt */
	custom.intena = IF_SETCLR | IF_PORTS;
    }
}

/*
 * DMA transfer mask for GVP Series II SCSI controller.
 * Some versions can only DMA into the 24 bit address space
 * (0->16M).  Others can DMA into the full 32 bit address
 * space.  The default is to only allow DMA into the 24 bit
 * address space.  The "gvp11=0xFFFFFFFE" setup parameter can
 * be supplied to force an alternate (32 bit) mask.
 */
static int gvp11_xfer_mask = GVP11_XFER_MASK;

void gvp11_setup (char *str, int *ints)
{
    gvp11_xfer_mask = ints[1];
}

static int dma_setup (Scsi_Cmnd *cmd, int dir_in)
{
    unsigned short cntr = GVP11_DMAC_INT_ENABLE;
    unsigned long addr = VTOP(cmd->SCp.ptr);

    /* don't allow DMA if the physical address is bad */
    if (addr & gvp11_xfer_mask ||
	(!dir_in && mm_end_of_chunk (addr, cmd->SCp.this_residual)))
    {

        uchar *bounce_buffer;
	unsigned int bounce_len = (cmd->SCp.this_residual + 511) & ~0x1ff;
	    
	/* FEF: try to grab a bounce buffer by scsi_malloc()
	 * If this doesn't work, try to chip_alloc it. if this fails
	 * either, return TRUE, no DMA possible for this transfer.
	 */
        
	bounce_buffer = scsi_malloc(bounce_len);

	if (bounce_buffer) {

	  /* FEF: seems we got a buffer from scsi_malloc(). Now check
	   * if DMA is possible in this region. If not, again try to
	   * use the "last chance exit" chip ram. If this fails, 
	   * return TRUE, no buffer, no DMA ;)
	   */

	  addr = VTOP(bounce_buffer);

	  if (addr & gvp11_xfer_mask) {
	    scsi_free(bounce_buffer, bounce_len);
	    bounce_buffer = NULL;
	  }
	}
        if (!bounce_buffer) {
	  bounce_buffer = amiga_chip_alloc(bounce_len);

	  if (!bounce_buffer) {
            HDATA(cmd->host)->dma_bounce_buffer = NULL;
            HDATA(cmd->host)->dma_bounce_len = 0;
	    return 1;
	  }
	  addr = ZTWO_PADDR(bounce_buffer);
	  HDATA(cmd->host)->dma_buffer_pool = BUF_CHIP_ALLOCED;
	} else 
	  HDATA(cmd->host)->dma_buffer_pool = BUF_SCSI_ALLOCED;

	/* FEF: Finally it turned out we have a buffer.
	 * Now all we have to do is fill the gaps within the
	 * hostdata structure and start DMA
	 */

	HDATA(cmd->host)->dma_bounce_buffer = bounce_buffer;
	HDATA(cmd->host)->dma_bounce_len = bounce_len;
	    
	/* copy to bounce buffer for a write */
	if (!dir_in) {
	  memcpy (bounce_buffer, cmd->SCp.ptr, cmd->SCp.this_residual);
	}
    }

    /* setup dma direction */
    if (!dir_in)
	cntr |= GVP11_DMAC_DIR_WRITE;

    HDATA(cmd->host)->dma_dir = dir_in;
    DMA(cmd->host)->CNTR = cntr;

    /* setup DMA *physical* address */
    DMA(cmd->host)->ACR = addr;

    if (dir_in)
	/* invalidate any cache */
	cache_clear (addr, cmd->SCp.this_residual);
    else
	/* push any dirty cache */
	cache_push (addr, cmd->SCp.this_residual);

    /* start DMA */
    DMA(cmd->host)->ST_DMA = 1;

    /* return success */
    return 0;
}

static void dma_stop (struct Scsi_Host *instance, Scsi_Cmnd *SCpnt,
		      int status)
{
    /* stop DMA */
    DMA(instance)->SP_DMA = 1;
    /* remove write bit from CONTROL bits */
    DMA(instance)->CNTR = GVP11_DMAC_INT_ENABLE;

    /* copy from a bounce buffer, if necessary */
    if (status && HDATA(instance)->dma_bounce_buffer) {

      if( HDATA(instance)->dma_dir )
	memcpy (SCpnt->SCp.ptr, 
		HDATA(instance)->dma_bounce_buffer,
		SCpnt->SCp.this_residual);

      if (HDATA(instance)->dma_buffer_pool == BUF_SCSI_ALLOCED) {
	scsi_free (HDATA(instance)->dma_bounce_buffer,
		   HDATA(instance)->dma_bounce_len);
      } else {
	amiga_chip_free(HDATA(instance)->dma_bounce_buffer);
      }
		     
      HDATA(instance)->dma_bounce_buffer = NULL;
      HDATA(instance)->dma_bounce_len = 0;
    }
}

int gvp11_detect(Scsi_Host_Template *tpnt)
{
    static unsigned char called = 0;
    struct Scsi_Host *instance;
    int num_gvp11 = 0;
    caddr_t address;
    enum GVP_ident epc;
    int key, index = 0;
    struct ConfigDev *cd;

    if (called)
	return 0;
    called = 1;

    while ((key = zorro_find(MANUF_GVP, PROD_GVPIISCSI, 0, index)) ||
	   (key = zorro_find(MANUF_GVP, PROD_GVPIISCSI_2, 0, index))) {
      index = key;
	cd = zorro_get_board(key);
	/* check extended product code */
	address = cd->cd_BoardAddr;
	epc = *(unsigned short *)ZTWO_VADDR(address + 0x8000);

	epc = epc & GVP_EPCMASK;

	/* 
	 * This should (hopefully) be the correct way to identify
	 * all the different GVP SCSI controllers (except for the
	 * SERIES I though).
	 */
	if (!((epc == GVP_A1291_SCSI) || 
	      (epc == GVP_GFORCE_040_SCSI) ||
	      (epc == GVP_GFORCE_030_SCSI) ||
	      (epc == GVP_A530_SCSI) ||
	      (epc == GVP_COMBO_R4_SCSI) ||
	      (epc == GVP_COMBO_R3_SCSI) ||
	      (epc == GVP_SERIESII)))
	    continue;

	instance = scsi_register (tpnt, sizeof (struct WD33C93_hostdata));
	instance->base = (unsigned char *)ZTWO_VADDR(address);
	DMA(instance)->secret2 = 1;
	DMA(instance)->secret1 = 0;
	DMA(instance)->secret3 = 15;
	while (DMA(instance)->CNTR & GVP11_DMAC_BUSY) ;
	DMA(instance)->CNTR = 0;
	wd33c93_init(instance, (wd33c93_regs *)&(DMA(instance)->SASR),
		     dma_setup, dma_stop, WD33C93_FS_8_10);
/*	host_nr++; */
	if (num_gvp11++ == 0) {
	    first_instance = instance;
	    gvp11_template = instance->hostt;
	    add_isr(IRQ_AMIGA_PORTS, gvp11_intr, 0, NULL, "gvp11 SCSI");
	}
	DMA(instance)->CNTR = GVP11_DMAC_INT_ENABLE;
	zorro_config_board(key, 0);
    }

    return num_gvp11;
}
